﻿using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Parsing;
using OpcLabs.DotProlog.Parsing;
using OpcLabs.DotProlog.Terms;

namespace DotPrologTest.Parsing
{
    [TestClass]
    public class TermParserTest
    {
        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void TryParseTest_IScanner_FunctorNotation()
        {
            Token[] tokens = 
            {
                Token.CreateName("animal"),
                Token.Create(TokenType.OpenCt),
                Token.CreateName("cat"),
                Token.Create(TokenType.Close),
                Token.Create(TokenType.End)
            };

            var target = new TermParser();
            var scanner = new Scanner<Token>(((IEnumerable<Token>)tokens).GetEnumerator());
            var parsingContext = new ParsingContext();
            TermParsingError error = target.TryParse(scanner, ParsingParameters.Default, parsingContext, out Term term);
            Assert.IsNull(error);
            Assert.IsNotNull(term);
            Assert.IsFalse(scanner.AtEnd);
            Token current = scanner.Current;
            Assert.IsNotNull(current);
            Assert.AreEqual(TokenType.End, current.TokenType);
            Assert.AreEqual(0, parsingContext.VariableNameDictionary.Count);

            Term expected = new NTerm("animal"){"cat"};
            Solution solution = TermUtilities.UnifyWithoutOccursCheck(term, expected);
            Assert.IsNotNull(solution);
            Assert.AreEqual(0, solution.Count);
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
