﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using JetBrains.Annotations;
using OpcLabs.BaseLib.Testing;

namespace OpcLabs.UnitTestLib.Testers
{
    public class BinarySerializationTester : Tester
    {
        public BinarySerializationTester()
        {
            TestCategories.Add("BinarySerialization");
            TestCategories.Add("Serialization");
        }

        public BinarySerializationTester([NotNull] Action<object, object> assertion) : this()
        {
            Assertion = assertion;
        }

        [CanBeNull]
        public Action<object, object> Assertion { get; }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            // ReSharper disable once LoopCanBePartlyConvertedToQuery
            foreach (KeyValuePair<string, CreatorEntry<InstancesAttribute>> instanceCreator in InstanceCreators)
            {
                string instanceName = instanceCreator.Key;
                Debug.Assert(instanceName != null);
                AddTestMethod(CreateRoundtripTestMethod(instanceName));
            }
        }

        [NotNull]
        private TestMethod CreateRoundtripTestMethod([NotNull] string instanceName)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Roundtrip_{0}", instanceName),
                () =>
                    {
                        object instance = CreateInstance(instanceName);
                        if (Assertion != null)
                            Tests.BinarySerializationTest(instance, Assertion);
                        else
                            Tests.BinarySerializationTest(instance);
                    });
        }
    }
}
