﻿using System;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.OperationModel;

namespace OpcLabs.BaseLibTest.OperationModel
{
    public partial class ValueResultTest
    {
        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void Formatting_F_Failure_ErrorId_Test()
        {
            Assert.AreEqual("*** Failure -2: Some error [...] (0 diagnostics); null {}", new ValueResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...] (0 diagnostics); null {}", new ValueResult { ErrorMessage = "Some error\nSecond line" }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success (4 diagnostics); 1.999 {System.Double}",
                new ValueResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList()),
                    Value = 1.999
                }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success (0 diagnostics); 1.999 {System.Double}", new ValueResult { Value = 1.999 }.ToString("F"));
        }

        [TestMethod]
        public void Formatting_F_Success_Null_Test()
        {
            Assert.AreEqual("Success (0 diagnostics); null {}", new ValueResult().ToString("F"));
        }

        [TestMethod]
        public void Formatting_g_Failure_ErrorId_Test()
        {
            Assert.AreEqual("*** Failure -2: Some error [...]", new ValueResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...]", new ValueResult { ErrorMessage = "Some error\nSecond line" }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success; 1.999",
                new ValueResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList()),
                    Value = 1.999
                }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success; 1.999", new ValueResult { Value = 1.999 }.ToString("g"));
        }

        [TestMethod]
        public void Formatting_g_Success_Null_Test()
        {
            Assert.AreEqual("Success; null", new ValueResult().ToString("g"));
        }

        [TestMethod]
        public void ToString_Failure_ErrorId_Test()
        {
            Assert.AreEqual("*** Failure -2: Some error [...]", new ValueResult { ErrorMessage = "Some error\nSecond line", ErrorId = "-2" }.ToString());
        }

        [TestMethod]
        public void ToString_Failure_NoErrorId_Test()
        {
            Assert.AreEqual("*** Failure System.Exception: Some error [...]", new ValueResult { ErrorMessage = "Some error\nSecond line" }.ToString());
        }

        [TestMethod]
        public void ToString_Success_Diagnostics_Test()
        {
            Assert.AreEqual(
                "Success (4 diagnostics); 1.999 {System.Double}",
                new ValueResult
                {
                    Diagnostics = new ExceptionCollection(Enumerable.Repeat<Exception>(new ApplicationException(), 4).ToList()),
                    Value = 1.999
                }.ToString());
        }

        [TestMethod]
        public void ToString_Success_NoDiagnostics_Test()
        {
            Assert.AreEqual("Success; 1.999 {System.Double}", new ValueResult { Value = 1.999 }.ToString());
        }

        [TestMethod]
        public void ToString_Success_Null_Test()
        {
            Assert.AreEqual("Success; null {}", new ValueResult().ToString());
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
