﻿
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.EasyOpc;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.AlarmsAndEvents
{
    public class AELiveTestEnvironment
    {
        #region Static Public Fields

        [NotNull]
        static public readonly AELiveTestEnvironment Current = new AELiveTestEnvironment();

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcLabs_KitEventServer OpcLabs_KitEventServer = new OpcLabs_KitEventServer(AELiveTestDefinitions.Locations.OpcComTest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<ServerDescriptor[]> OpcLabs_KitEventServer_OptimisticSupportedServerDescriptorDataRows
        {
            get
            {
                return OpcLabs_KitEventServer.AllServerDescriptors
                    //.Where(IsSupportedOptimistic)
                    .Select(descriptor => new[] { descriptor });
            }
        }

        #endregion



        #region Public Methods

        public bool CanConnectServer([NotNull] ServerDescriptor serverDescriptor)
        {
            return TestServer(serverDescriptor) == 0;
        }

        public void PrecheckServer([NotNull] ServerDescriptor serverDescriptor)
        {
            AssertUtilities.Phase($"Pre-checking server \"{serverDescriptor}\"");

            int error = TestServer(serverDescriptor);
            if (error != 0)
            {
                string message = $"Server \"{serverDescriptor}\" pre-check failed: {error}";
                AssertUtilities.WriteLine(message);
                Assert.Inconclusive(message);
            }
        }

        public int TestServer([NotNull] ServerDescriptor serverDescriptor)
        {
            int error;
            lock (_testServerCache)
            {
                if (!_testServerCache.TryGetValue(serverDescriptor, out error))
                {
                    error = InternalTestServer(serverDescriptor);
                    _testServerCache.AddNew(serverDescriptor, error);
                }
            }
            return error;
        }

        #endregion



        #region Private Methods

        private int InternalTestServer([NotNull] ServerDescriptor serverDescriptor)
        {
            switch (serverDescriptor.Technology)
            {
                case OpcTechnologies.Com:
// TODO:                    if (!serverDescriptor.CanInstantiate())
// TODO:                        return -1;
                    return 0;

                /* TODO:

                string machineName = serverDescriptor.MachineName;
                if (!machineName.IsEmpty())
                {
                    PingReply pingReply;
                    try
                    {
                        pingReply = PingUtilities.Send(machineName);
                    }
                    catch (NotSupportedException)
                    {
                        // address is an IPv6 address and the local computer is running an operating system earlier than Windows 2000. 
                        return -2;
                    }
                    catch (PingException)
                    {
                        // An exception was thrown while sending or receiving the ICMP messages. See the inner exception for the exact 
                        // exception that was thrown.
                        return -1;
                    }
                    if (pingReply.Status != IPStatus.Success)
                        return -3;
                }

                // ServerDescriptor pre-check passes for all pingable appear-remote machines, regardless of the server class.
                bool appearsLocal =
                    machineName.IsEmpty() ||
                    HostNameUtilities.IsLocalHostOrAddress(machineName) ||
                    (machineName.ToUpper() == System.Environment.MachineName.ToUpper()) ||
                    HostNameUtilities.Match(machineName, System.Net.Dns.GetHostName());
                // Cannot check instantiation feasibility on remote machines.
                if (appearsLocal)
                {
                    ServerDescriptor localServerDescriptor = serverDescriptor.TypedClone();
                    localServerDescriptor.MachineName = "";
                    if (!localServerDescriptor.CanInstantiate())
                        return -4;
                }

                return 0;
                */

                case OpcTechnologies.Xml:
                    return TcpClientUtilities.TestCreate(serverDescriptor, millisecondsTimeout: 1500);
            }

            return 0;
        }

        #endregion



        #region Private Fields

        [NotNull]
        private readonly Dictionary<ServerDescriptor, int> _testServerCache = new Dictionary<ServerDescriptor, int>();

        #endregion
    }
}
