﻿
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using JetBrains.Annotations;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Persistence;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;

namespace OpcLabs.EasyOpcLiveTest.DataAccess.Environment
{
    // ReSharper disable once InconsistentNaming
    public class OpcFoundation_XmlDaSampleServer : CommonServer
    {
        #region Public Types

        public class ServerDescriptorsInfo
        {
            public ServerDescriptorsInfo([NotNull] OpcFoundation_XmlDaSampleServer parent)
            {
                _parent = parent;
            }

            [NotNull]
            public ServerDescriptor Http => new ServerDescriptor(_parent._endpointUrlStrings.Http);

            [NotNull]
            private readonly OpcFoundation_XmlDaSampleServer _parent;
        }

        #endregion



        #region Private Types

        private class EndpointUrlStringsInfo
        {
            public EndpointUrlStringsInfo([NotNull] OpcFoundation_XmlDaSampleServer parent)
            {
                _parent = parent;
            }

            [NotNull]
            public string Http => "http://" + _parent.Location + "/XmlDaSampleServer/Service.asmx";

            [NotNull]
            private readonly OpcFoundation_XmlDaSampleServer _parent;
        }

        #endregion



        #region Public Constructors

        public OpcFoundation_XmlDaSampleServer([NotNull] string location) 
            : base(location)
        {
            ServerDescriptors = new ServerDescriptorsInfo(this);
            _endpointUrlStrings = new EndpointUrlStringsInfo(this);
        }

        #endregion



        #region Static Public Properties

        [NotNull]
        static public IEnumerable<DAItemCatalogEntry> AllItemCatalogEntries
        {
            get
            {
                // /Dynamic/Analog Types

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Fools",
                    browsePath: "/Dynamic/Analog Types/Fools",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Fools",
                    isBranch: true,
                    isLeaf: true,
                    hasChildren: true))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I2,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int",
                    browsePath: "/Dynamic/Analog Types/Int",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Int",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double",
                    browsePath: "/Dynamic/Analog Types/Double",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Double",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.R8,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int[]",
                    browsePath: "/Dynamic/Analog Types/Int[]",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Int[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.ArrayOfI4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double[]",
                    browsePath: "/Dynamic/Analog Types/Double[]",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Double[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.ArrayOfR8,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                // /Dynamic/Analog Types/Fools

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Rosencrantz",
                    browsePath: "/Dynamic/Analog Types/Fools/Rosencrantz",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Fools/Rosencrantz",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Guildenstern",
                    browsePath: "/Dynamic/Analog Types/Fools/Guildenstern",
                    nodePath: "",
                    itemId: "Dynamic/Analog Types/Fools/Guildenstern",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                // /Dynamic/Enumerated Types

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Gems",
                    browsePath: "/Dynamic/Enumerated Types/Gems",
                    nodePath: "",
                    itemId: "Dynamic/Enumerated Types/Gems",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Fellowship",
                    browsePath: "/Dynamic/Enumerated Types/Fellowship",
                    nodePath: "",
                    itemId: "Dynamic/Enumerated Types/Fellowship",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 1000,
                    SlowestRateOfChange = 1000,
                };

                // /Static/Analog Types

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int",
                    browsePath: "/Static/Analog Types/Int",
                    nodePath: "",
                    itemId: "Static/Analog Types/Int",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double",
                    browsePath: "/Static/Analog Types/Double",
                    nodePath: "",
                    itemId: "Static/Analog Types/Double",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.R8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int[]",
                    browsePath: "/Static/Analog Types/Int[]",
                    nodePath: "",
                    itemId: "Static/Analog Types/Int[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.ArrayOfI4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double[]",
                    browsePath: "/Static/Analog Types/Double[]",
                    nodePath: "",
                    itemId: "Static/Analog Types/Double[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.ArrayOfR8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                // /Static/ArrayTypes

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Byte[]",
                    browsePath: "/Static/ArrayTypes/Byte[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Byte[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfUI1,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Byte[]",
                    browsePath: "/Static/ArrayTypes/Byte[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Byte[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfUI1,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Short[]",
                    browsePath: "/Static/ArrayTypes/Short[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Short[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfI2,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "UShort[]",
                    browsePath: "/Static/ArrayTypes/UShort[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/UShort[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfUI2,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int[]",
                    browsePath: "/Static/ArrayTypes/Int[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Int[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfI4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "UInt[]",
                    browsePath: "/Static/ArrayTypes/UInt[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/UInt[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfUI4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Long[]",
                    browsePath: "/Static/ArrayTypes/Long[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Long[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = ((VarType)VarTypes.I8).MakeArrayType(), // IMPROVE: investigate and document why and how this works
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "ULong[]",
                    browsePath: "/Static/ArrayTypes/ULong[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/ULong[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = ((VarType)VarTypes.UI8).MakeArrayType(), // IMPROVE: investigate and document why and how this works
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Float[]",
                    browsePath: "/Static/ArrayTypes/Float[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Float[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfR4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double[]",
                    browsePath: "/Static/ArrayTypes/Double[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Double[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfR8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Decimal[]",
                    browsePath: "/Static/ArrayTypes/Decimal[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Decimal[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfDecimal,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "DateTime[]",
                    browsePath: "/Static/ArrayTypes/DateTime[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/DateTime[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfDate,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Boolean[]",
                    browsePath: "/Static/ArrayTypes/Boolean[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/Boolean[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfBool,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "String[]",
                    browsePath: "/Static/ArrayTypes/String[]",
                    nodePath: "",
                    itemId: "Static/ArrayTypes/String[]",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.ArrayOfBStr,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                // /Static/Enumerated Types

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Gems",
                    browsePath: "/Static/Enumerated Types/Gems",
                    nodePath: "",
                    itemId: "Static/Enumerated Types/Gems",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Fellowship",
                    browsePath: "/Static/Enumerated Types/Fellowship",
                    nodePath: "",
                    itemId: "Static/Enumerated Types/Fellowship",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Unknown,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                // /Static/Limited Access

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Read Only 1",
                    browsePath: "/Static/Limited Access/Read Only 1",
                    nodePath: "",
                    itemId: "Static/Limited Access/Read Only 1",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Readable,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Read Only 2",
                    browsePath: "/Static/Limited Access/Read Only 2",
                    nodePath: "",
                    itemId: "Static/Limited Access/Read Only 2",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Readable,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.R8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Write Only 1",
                    browsePath: "/Static/Limited Access/Write Only 1",
                    nodePath: "",
                    itemId: "Static/Limited Access/Write Only 1",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Writable,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Write Only 2",
                    browsePath: "/Static/Limited Access/Write Only 2",
                    nodePath: "",
                    itemId: "Static/Limited Access/Write Only 2",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Writable,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.R8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                // /Static/Simple Types

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Byte",
                    browsePath: "/Static/Simple Types/Byte",
                    nodePath: "",
                    itemId: "Static/Simple Types/Byte",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.UI1,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "SByte",
                    browsePath: "/Static/Simple Types/SByte",
                    nodePath: "",
                    itemId: "Static/Simple Types/SByte",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I1,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Short",
                    browsePath: "/Static/Simple Types/Short",
                    nodePath: "",
                    itemId: "Static/Simple Types/Short",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I2,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "UShort",
                    browsePath: "/Static/Simple Types/UShort",
                    nodePath: "",
                    itemId: "Static/Simple Types/UShort",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.UI2,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Int",
                    browsePath: "/Static/Simple Types/Int",
                    nodePath: "",
                    itemId: "Static/Simple Types/Int",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "UInt",
                    browsePath: "/Static/Simple Types/UInt",
                    nodePath: "",
                    itemId: "Static/Simple Types/UInt",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.UI4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Long",
                    browsePath: "/Static/Simple Types/Long",
                    nodePath: "",
                    itemId: "Static/Simple Types/Long",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.I8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "ULong",
                    browsePath: "/Static/Simple Types/ULong",
                    nodePath: "",
                    itemId: "Static/Simple Types/ULong",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.UI8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Float",
                    browsePath: "/Static/Simple Types/Float",
                    nodePath: "",
                    itemId: "Static/Simple Types/Float",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.R4,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Double",
                    browsePath: "/Static/Simple Types/Double",
                    nodePath: "",
                    itemId: "Static/Simple Types/Double",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.R8,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Decimal",
                    browsePath: "/Static/Simple Types/Decimal",
                    nodePath: "",
                    itemId: "Static/Simple Types/Decimal",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.Decimal,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "DateTime",
                    browsePath: "/Static/Simple Types/DateTime",
                    nodePath: "",
                    itemId: "Static/Simple Types/DateTime",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.Date,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "Boolean",
                    browsePath: "/Static/Simple Types/Boolean",
                    nodePath: "",
                    itemId: "Static/Simple Types/Boolean",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.Bool,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };

                yield return new DAItemCatalogEntry(new DANodeElement(
                    name: "String",
                    browsePath: "/Static/Simple Types/String",
                    nodePath: "",
                    itemId: "Static/Simple Types/String",
                    isBranch: false,
                    isLeaf: true,
                    hasChildren: false))
                {
                    PropertyCatalog = CommonPropertyCatalogEntries,
                    AccessRights = DAAccessRights.Both,
                    BehaviorKind = OpcItemBehaviorKind.Register,
                    DataType = VarTypes.BStr,
                    FastestRateOfChange = 0,
                    SlowestRateOfChange = Timeout.Infinite,
                };
            }
        }

        [NotNull]
        static public IEnumerable<DAItemCatalogEntry[]> AllItemCatalogEntryDataRows => AllItemCatalogEntries.Select(entry => new[] { entry });

        [NotNull]
        static public IEnumerable<DAPropertyCatalogEntry[]> CommonPropertyCatalogEntryDataRows => CommonPropertyCatalogEntries.Select(entry => new[] { entry });

        [NotNull]
        static public IEnumerable<DAItemCatalogEntry[]> NonReadableItemCatalogEntryDataRows => AllItemCatalogEntries.NonReadables().Select(entry => new[] { entry });

        [NotNull]
        static public IEnumerable<DAItemCatalogEntry[]> ReadableItemCatalogEntryDataRows => AllItemCatalogEntries.Readables().Select(entry => new[] { entry });

        [NotNull]
        static public IEnumerable<DAItemCatalogEntry[]> ReadWriteableItemCatalogEntryDataRows => AllItemCatalogEntries.ReadWriteables().Select(entry => new[] { entry });

        #endregion



        #region Static Private Properties

        [NotNull]
        static private IEnumerable<DAPropertyCatalogEntry> CommonPropertyCatalogEntries
        {
            get
            {
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.DataType, "highEU",
                    VarTypes.I2, "Item Canonical Data Type"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Value, "value",
                    VarTypes.Empty, "Item Value"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Quality, "quality",
                    VarTypes.I2, "Item Quality"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.Timestamp, "timestamp",
                    VarTypes.Date, "Item Timestamp"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.AccessRights,
                    "accessRights",
                    VarTypes.I4, "Item Access Rights"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.ScanRate, "scanRate",
                    VarTypes.R4, "Server Scan Rate"));
                yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.EUType, "euType",
                    VarTypes.I4, "Item EU Type"));

                //yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.HighEU, VarTypes.R8, "High EU"));
                //yield return new DAPropertyCatalogEntry(new DAPropertyElement(DAPropertyIds.LowEU, VarTypes.R8, "Low EU"));

                // TODO: have more
            }
        }

        #endregion



        #region Public Property

        [NotNull]
        public ServerDescriptorsInfo ServerDescriptors { get; }

        public override IEnumerable<ServerDescriptor> AllServerDescriptors
        {
            get
            {
                yield return ServerDescriptors.Http;
            }
        }

        #endregion



        #region Private Fields

        [NotNull]
        private readonly EndpointUrlStringsInfo _endpointUrlStrings;

        #endregion
    }
}
