﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.OperationModel;

namespace OpcLabs.EasyOpcLiveTest.DataAccess.Extensions
{
    public abstract class EasyDAClientExtensionTestBase
    {
        /// <summary>
        ///Gets or sets the test context which provides
        ///information about and functionality for the current test run.
        ///</summary>
        public TestContext TestContext { get; set; }



        /// <summary>
        ///A test for GetAccessRightsPropertyValue
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetAccessRightsPropertyValueTest()
        {
            IEasyDAClient client = CreateClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4";
            DAAccessRight expected = DAAccessRights.Both;
            DAAccessRight actual;
            try
            {
                actual = client.GetAccessRightsPropertyValue(machineName, serverClass, itemId);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            Assert.AreEqual(expected, actual);
        }

        /// <summary>
        ///A test for ReadItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueArrayOfInt32Test()
        {
            IEasyDAClient client = CreateClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.ReadValue_ArrayOfI4";
            int[] actual;
            try
            {
                actual = client.ReadItemValueArrayOfInt32(machineName, serverClass, itemId);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            int expectedElement = actual[actual.GetLowerBound(0)];
            for (int i = actual.GetLowerBound(0) + 1; i <= actual.GetUpperBound(0); i++)
                Assert.AreEqual(expectedElement, actual[i]);
        }

        /// <summary>
        ///A test for ReadItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueArrayOfInt32Test1()
        {
            IEasyDAClient client = CreateClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.ReadValue_ArrayOfI4";
            int[] actual;
            try
            {
                actual = client.ReadItemValueArrayOfInt32(serverDescriptor, itemId);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            int expectedElement = actual[actual.GetLowerBound(0)];
            for (int i = actual.GetLowerBound(0) + 1; i <= actual.GetUpperBound(0); i++)
                Assert.AreEqual(expectedElement, actual[i]);
        }

        /// <summary>
        ///A test for ReadItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueInt32Test()
        {
            const int value = 892736297;
            IEasyDAClient client = CreateClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4"; // IMPROVE: Use an item with guaranteed constant value instead
            try
            {
                client.WriteItemValue(machineName, serverClass, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            const int expected = value;
            int actual = client.ReadItemValueInt32(machineName, serverClass, itemId);
            Assert.AreEqual(expected, actual);
        }

        /// <summary>
        ///A test for ReadItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueInt32Test1()
        {
            const int value = 892736297;
            IEasyDAClient client = CreateClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_I4"; // IMPROVE: Use an item with guaranteed constant value instead
            try
            {
                client.WriteItemValue<int>(serverDescriptor, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            const int expected = value;
            int actual = client.ReadItemValueInt32(serverDescriptor, itemId);
            Assert.AreEqual(expected, actual);
        }

        /// <summary>
        ///A test for WriteItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueArrayOfInt32Test()
        {
            IEasyDAClient client = CreateClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_ArrayOfI4";
            int[] value = { 1, 2, 3 };
            try
            {
                client.WriteItemValueArrayOfInt32(machineName, serverClass, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueArrayOfInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueArrayOfInt32Test1()
        {
            IEasyDAClient client = CreateClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_ArrayOfI4";
            int[] value = { 1, 2, 3 };
            try
            {
                client.WriteItemValueArrayOfInt32(serverDescriptor, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueInt32Test()
        {
            IEasyDAClient client = CreateClient();
            const string machineName = "";
            const string serverClass = "OPCLabs.KitServer.2";
            const string itemId = "Simulation.Register_I4";
            const int value = 0; // TODO: Initialize to an appropriate value
            try
            {
                client.WriteItemValueInt32(machineName, serverClass, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            // IMPROVE: Test that is has been written correctly
        }

        /// <summary>
        ///A test for WriteItemValueInt32
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueInt32Test1()
        {
            IEasyDAClient client = CreateClient();
            var serverDescriptor = new ServerDescriptor("OPCLabs.KitServer.2");
            const string itemId = "Simulation.Register_I4";
            const int value = 0; // TODO: Initialize to an appropriate value
            try
            {
                client.WriteItemValueInt32(serverDescriptor, itemId, value);
            }
            catch (OpcException opcException)
            {
                if (!client.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            // IMPROVE: Test that is has been written correctly
        }


        [NotNull]
        protected abstract IEasyDAClient CreateClient();
    }
}
