﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Identification;
using OpcLabs.BaseLib.Instrumentation.Internal;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Implementation;
using OpcLabs.EasyOpc.DataAccess.Internal;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.DataAccess.Optimization.Implementation;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.Implementation
{
    [TestClass]
    public class SubscriptionBlendingEasyDAClientTest
    {
        [TestMethod]
        public void Test()
        {
            TestSection.Arrange();

            var apiMonitoringEntryList = new List<ApiMonitoringEntry>();

            // ReSharper disable once ParameterOnlyUsedForPreconditionCheck.Local
            void MonitorApiHandler(object sender, DataEventArgs<ApiMonitoringEntry> eventArgs, int level)
            {
                Assert.IsNotNull(sender);
                Assert.IsNotNull(eventArgs);
                ApiMonitoringEntry apiMonitoringEntry = eventArgs.Data;
                Assert.IsNotNull(apiMonitoringEntry);
                AssertUtilities.WriteLine($"({apiMonitoringEntryList.Count}) {new String(' ', level * 4)}{apiMonitoringEntry}");
                apiMonitoringEntryList.Add(apiMonitoringEntry);
            }

            var simulatedClient = new SimulatedEasyDAClient();
            var simulatedClientApiMonitoringClient = new ApiMonitoringEasyDAClient(simulatedClient, Luid.NewLuid());
            simulatedClientApiMonitoringClient.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 1);

            var subscriptionBlendingClient = new SubscriptionBlendingEasyDAClient(simulatedClientApiMonitoringClient);
            var target = new ApiMonitoringEasyDAClient(subscriptionBlendingClient, Luid.NewLuid());
            target.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 0);
            target.ItemChanged += (sender, args) => AssertUtilities.WriteLine(args);

            TestSection.Act();

            int requestedUpdateRate1A = 1000;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Subscribe to item 1 subscription A, requested update rate {requestedUpdateRate1A}.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments1A = new EasyDAItemSubscriptionArguments("s1", "i1", requestedUpdateRate1A, "ItemSubscription1A");
            int handle1A = target.SubscribeItem(itemSubscriptionArguments1A);
            AssertUtilities.WriteLine($"{nameof(handle1A)}: {handle1A}");
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0]).Length);
            Assert.AreEqual(itemSubscriptionArguments1A.ServerDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ServerDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1A.ItemDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ItemDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1A.GroupParameters, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[6].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[6].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[6].IsReturn);
            int innerHandle1A = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle1A)}: {innerHandle1A}");
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate1B = 1000;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Subscribe to item 1 subscription B, requested update rate {requestedUpdateRate1B}.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments1B = new EasyDAItemSubscriptionArguments("s1", "i1", requestedUpdateRate1B, "ItemSubscription1B");
            int handle1B = target.SubscribeItem(itemSubscriptionArguments1B);
            AssertUtilities.WriteLine($"{nameof(handle1B)}: {handle1B}");
            // ReSharper disable PossibleNullReferenceException
            foreach (ApiMonitoringEntry apiMonitoringEntry in apiMonitoringEntryList)
                Assert.AreNotSame(simulatedClient, apiMonitoringEntry.TargetOrSourceObject);
            Assert.AreSame(subscriptionBlendingClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(target, apiMonitoringEntryList[1].Parameters[0]);
            Assert.AreEqual(itemSubscriptionArguments1B, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[1].Parameters[1]).Arguments);
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate1C = 2000;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Subscribe to item 1 subscription C, requested update rate {requestedUpdateRate1C}.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments1C = new EasyDAItemSubscriptionArguments("s1", "i1", requestedUpdateRate1C, "ItemSubscription1C");
            int handle1C = target.SubscribeItem(itemSubscriptionArguments1C);
            AssertUtilities.WriteLine($"{nameof(handle1C)}: {handle1C}");
            // ReSharper disable PossibleNullReferenceException
            foreach (ApiMonitoringEntry apiMonitoringEntry in apiMonitoringEntryList)
                Assert.AreNotSame(simulatedClient, apiMonitoringEntry.TargetOrSourceObject);
            Assert.AreSame(subscriptionBlendingClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreSame(target, apiMonitoringEntryList[1].Parameters[0]);
            Assert.AreEqual(itemSubscriptionArguments1C, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[1].Parameters[1]).Arguments);
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate1D = 500;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Subscribe to item 1 subscription D, requested update rate {requestedUpdateRate1D}.");
            apiMonitoringEntryList.Clear();
            var groupParameters1D = new DAGroupParameters(requestedUpdateRate1D);
            EasyDAItemSubscriptionArguments itemSubscriptionArguments1D = new EasyDAItemSubscriptionArguments("s1", "i1", groupParameters1D, "ItemSubscription1D");
            int handle1D = target.SubscribeItem(itemSubscriptionArguments1D);
            AssertUtilities.WriteLine($"{nameof(handle1D)}: {handle1D}");
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[3].TypedMemberName);
            Assert.AreEqual(1, ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0]).Length);
            var actualHandleGroupArguments1D = ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0])[0];
            Assert.IsNotNull(actualHandleGroupArguments1D);
            Assert.AreEqual(innerHandle1A, actualHandleGroupArguments1D.Handle);
            Assert.AreEqual(groupParameters1D, actualHandleGroupArguments1D.GroupParameters);
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate2A = 1000;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Subscribe to item 2, requested update rate {requestedUpdateRate2A}.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments2A = new EasyDAItemSubscriptionArguments("s1", "i2", requestedUpdateRate1C, "ItemSubscription2A");
            int handle2A = target.SubscribeItem(itemSubscriptionArguments2A);
            AssertUtilities.WriteLine($"{nameof(handle2A)}: {handle2A}");
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0]).Length);
            Assert.AreEqual(itemSubscriptionArguments2A.ServerDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ServerDescriptor);
            Assert.AreEqual(itemSubscriptionArguments2A.ItemDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].ItemDescriptor);
            Assert.AreEqual(itemSubscriptionArguments2A.GroupParameters, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[1].Parameters[0])[0].GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[6].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[6].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[6].IsReturn);
            int innerHandle2A = ((int[])apiMonitoringEntryList[6].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle2A)}: {innerHandle2A}");
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate1A2 = 3000;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Change item 1 subscription A, new requested update rate {requestedUpdateRate1A2}.");
            apiMonitoringEntryList.Clear();
            var groupParameters1A2 = new DAGroupParameters(requestedUpdateRate1A2);
            target.ChangeItemSubscription(handle1A, groupParameters1A2);
            // ReSharper disable PossibleNullReferenceException
            foreach (ApiMonitoringEntry apiMonitoringEntry in apiMonitoringEntryList)
                Assert.AreNotSame(simulatedClient, apiMonitoringEntry.TargetOrSourceObject);
            // ReSharper restore PossibleNullReferenceException

            int requestedUpdateRate1A3 = 400;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Change item 1 subscription A, new requested update rate {requestedUpdateRate1A3}.");
            apiMonitoringEntryList.Clear();
            var groupParameters1A3 = new DAGroupParameters(requestedUpdateRate1A3);
            target.ChangeItemSubscription(handle1A, groupParameters1A3);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[3].TypedMemberName);
            Assert.AreEqual(1, ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0]).Length);
            var actualHandleGroupArguments1A3 = ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0])[0];
            Assert.IsNotNull(actualHandleGroupArguments1A3);
            Assert.AreEqual(innerHandle1A, actualHandleGroupArguments1A3.Handle);
            Assert.AreEqual(groupParameters1A3, actualHandleGroupArguments1A3.GroupParameters);
            // ReSharper restore PossibleNullReferenceException

            int locale1A4 = 1024;
            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine($"Change item 1 subscription A, new locale {locale1A4}.");
            apiMonitoringEntryList.Clear();
            var groupParameters1A4 = new DAGroupParameters(requestedUpdateRate1A3) {Locale = locale1A4};
            target.ChangeItemSubscription(handle1A, groupParameters1A4);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ChangeMultipleItemSubscriptions", apiMonitoringEntryList[3].TypedMemberName);
            Assert.AreEqual(1, ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0]).Length);
            var actualHandleGroupArguments1A4 = ((DAHandleGroupArguments[])apiMonitoringEntryList[3].Parameters[0])[0];
            Assert.IsNotNull(actualHandleGroupArguments1A4);
            Assert.AreEqual(innerHandle1A, actualHandleGroupArguments1A4.Handle);
            Assert.AreEqual(groupParameters1D, actualHandleGroupArguments1A4.GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[5].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[5].TypedMemberName);
            Assert.AreEqual(1, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[5].Parameters[0]).Length);
            Assert.AreEqual(itemSubscriptionArguments1A.ServerDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[5].Parameters[0])[0].ServerDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1A.ItemDescriptor, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[5].Parameters[0])[0].ItemDescriptor);
            Assert.AreEqual(groupParameters1A4, ((EasyDAItemSubscriptionArguments[])apiMonitoringEntryList[5].Parameters[0])[0].GroupParameters);
            Assert.AreSame(subscriptionBlendingClient, apiMonitoringEntryList[7].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ItemChanged", apiMonitoringEntryList[7].TypedMemberName);
            Assert.AreSame(target, apiMonitoringEntryList[7].Parameters[0]);
            Assert.AreEqual(itemSubscriptionArguments1A.ServerDescriptor, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[7].Parameters[1]).Arguments.ServerDescriptor);
            Assert.AreEqual(itemSubscriptionArguments1A.ItemDescriptor, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[7].Parameters[1]).Arguments.ItemDescriptor);
            Assert.AreEqual(groupParameters1A4, ((EasyDAItemChangedEventArgs)apiMonitoringEntryList[7].Parameters[1]).Arguments.GroupParameters);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[10].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.SubscribeMultipleItems", apiMonitoringEntryList[10].TypedMemberName);
            Assert.IsTrue(apiMonitoringEntryList[10].IsReturn);
            int innerHandle1A4 = ((int[])apiMonitoringEntryList[10].ReturnValue)[0];
            AssertUtilities.WriteLine($"{nameof(innerHandle1A4)}: {innerHandle1A4}");
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Unsubscribe from item 1 subscription A.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle1A);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[3].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.UnsubscribeMultipleItems", apiMonitoringEntryList[3].TypedMemberName);
            var actualInnerHandles1A = (IEnumerable<int>)apiMonitoringEntryList[3].Parameters[0];
            Assert.IsNotNull(actualInnerHandles1A);
            Assert.IsTrue(actualInnerHandles1A.SequenceEqual(new[] { innerHandle1A4 }));
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Unsubscribe from item 1 subscription B.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeItem(handle1B);
            // ReSharper disable PossibleNullReferenceException
            foreach (ApiMonitoringEntry apiMonitoringEntry in apiMonitoringEntryList)
                Assert.AreNotSame(simulatedClient, apiMonitoringEntry.TargetOrSourceObject);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Unsubscribe from all items.");
            apiMonitoringEntryList.Clear();
            target.UnsubscribeAllItems();
        }
    }
}
